# -*- coding: utf-8 -*-
"""
Created on Mon Jun 15 17:52:27 2020

@author: bugra
"""


"""Local adaptive Otsu thresholding.

This module performs local adaptive otsu thresholding on 3D volumetric images.


Classes: 
    
    otsu: a class to calculate local and global otsu thresholds for an image.

Functions:
    
    histogram: a function to generate the histogram of a greyscale image.
    
    otsu_threshold: a function to calculate a single, global otsu threshold for an image.
    
    get_preliminary_mask: a convenience function to select a subset of voxels after oversegmenting an image.
    
    _local_otsu: a function to calculate and store local otsu thresholds for an image, using arbitrary local windows
    """

import numpy as np
from numba import jit
from skimage import morphology

######################################################################################################################
############################################ Local adaptive otsu thresholding ########################################
######################################################################################################################


@jit(nopython=True)
def histogram(image, nbins=256):
    """This function generates the histogram of a greyscale image.
    
    Parameters:
    ----------------
        image: array 
                The input data, over which the histogram is calculated.
        nbins: int
                The number of equal-sized bins to calculate the histogram.
    
    Returns: 
    ----------------
        hist: array of dtype int
                The number of voxels corresponding to each bin.
        bin_centers: array of dtype float
                The grey values corresponding to the bin centers."""
    image = image.flatten()
    hist, bin_edges = np.histogram(image, bins=nbins, range=None)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.
    return (hist, bin_centers)

@jit(nopython=True)
def otsu_threshold(image, nbins=256):
    """This function calculates the otsu threshold of an image.
    
    Parameters: 
    ----------------
        image: array
                The input data, for which the otsu threshold is calculated. Can be n-dimensional.
        nbins: int
                The number of equal-sized bins used for the histogram calculation.
    
    Returns: 
    ----------------
        threshold: float
                The otsu threshold of the image."""
    hist, bin_centers = histogram(image.ravel(), nbins)
    hist = hist/1.0
    weight1 = np.cumsum(hist)
    weight2 = np.cumsum(hist[::-1])[::-1]
    mean1 = np.cumsum(hist * bin_centers) / weight1
    mean2 = (np.cumsum((hist * bin_centers)[::-1]) / weight2[::-1])[::-1]
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2
    idx = np.argmax(variance12)
    threshold = bin_centers[:-1][idx]
    return (threshold)

def get_preliminary_mask(image,threshold_factor=0.66,padder=0):
    """This function generates a binary mask on the input image based on otsu threshold multiplied with a factor.
    
    Parameters: 
    ----------------
        image: array
                The input data, on which the mask is generated. Can be n-dimensional.
        threshold_factor: float
                A factor that is used to tune the treshold height for generation of the preliminary mask. 
                This parameter is multiplied with the otsu threshold of the image to get a final threshold that 
                is used to binarise the image. Since this step is intended to oversegment the image, this parameter 
                should be typically between 0 and 1. If this parameter is chosen to be negative and the padder is chosen
                to be zero, then all voxels in the image will be in the final binary mask that is returned.
        padder: int
                The width of the mask borders to be assigned with empty voxels. For example, if padder is 3, a width of
                3 voxels from each border of the binary mask will be automatically assigned with False. This parameter
                ensures that the mask excludes any voxels too close to the borders so that the sliding window remains fully
                within the image frame during the local otsu implementation. 
    Returns: 
    ----------------
        coords: array of dtype int
                Coordinates of voxels included in the mask.
        mask: array of dtype bool
                Array that has the same number and size of dimensions as the image. 
        global_otsu: float 
                Global otsu threshold calculated on the image, multiplied with the 'threshold_factor'."""
    global_otsu=otsu_threshold(image)*threshold_factor
    mask=image>global_otsu
    if padder>0:
        mask[:padder,:padder,:padder]=False
        mask[-padder:,-padder:,-padder:]=False
    coords=np.argwhere(mask)
    return(coords,mask,global_otsu)
    
@jit(nopython=True)
def _local_otsu(image,coords,radii):
    """This function generates local thresholds by applying otsu method locally to chunks of volume.
    
    Parameters: 
    ----------------
        image: array
                The input data, over which the local otsu thresholds will be calculated. This array must be 3-dimensional.
        coords: array of dtype int
                Coordinates of voxels that represent the centers for the sliding window. The sliding window is 
                set around each coordinate to determine the local volume for threshold calculation. The otsu threshold
                is calculated for each chunk of volume and stored in a list. 
        radii: (int, int, int)
                Lengths of the sliding window radii, in the order of depth, row, col. 
    Returns: 
    ----------------
        threshlist: list
                The list of the calculated local thresholds. This has the same length as the parameter 'coords'."""
    threshlist=[]
    ir,jr,kr = radii
    for it in range(len(coords)):
        i,j,k=coords[it]
        location=image[i-ir:i+ir+1,j-jr:j+jr+1,k-kr:k+kr+1]
        if np.count_nonzero(location)==0:
            thresh=0
        else:
            thresh=otsu_threshold(location)
        threshlist.append(thresh)
    return(threshlist)

class otsu:
    """This class calculates local and global otsu thresholds and segments images using them.
    
    Attributes: 
    ----------------
        img: array
                The input data, over which the thresholds are calculated. This array must be 3-dimensional.
        threshold_factor: float
                a factor multiplied with otsu threshold of the image to determine the global threshold 
                used for obtaining a preliminary coarse mask. See also the 'threshold_factor' parameter 
                of the 'get_preliminary_mask' function in this file
        global_threshold: float
                A threshold that is smaller than otsu threshold of the image, used for obtaining a preliminary, 
                coarse mask.
        global_thresholded: array of dtype bool
                The preliminary binary mask obtained by applying the global threshold to the image.
        padder: int
                The number of empty voxels to be added to each border of the image. This parameter also determines the 
                width of the mask borders to be assigned with empty voxels. See also the 'padder' parameter of the 
                'get_preliminary_mask' function in this file.
        padded: array
                The input data padded with the 'padder' attribute to guarantee that the sliding window remains within the
                image borders during local thresholding.
        selected_coords: array of dtype int
                Coordinates of voxels included in the preliminary mask. These are the voxels sampled by the '_local_otsu'
                function. See also the 'coords' parameter of the '_local_otsu' function in this file
        radii: (int, int, int)
                Lengths of the sliding window radii, in the order of depth, row, col. See also the 'radii' 
                parameter of the '_local_otsu' function in this file.
        local_thresholds: array of dtype float
                Local thresholds mapped to an array with the same shape as 'img'
        local_thresholded: array of dtype bool
                Binary mask that is the result of thresholding the 'img' with both local and global thresholds
        local_thresholded_cleaned: array of dtype bool
                Binary mask that is the result of thresholding the 'img' with both local and global thresholds followed by
                a cleaning process that involves removal of small objects and border artifacts
        thresholding_status: int
                A reporter that indicates whether the method '_calculate_local_thresholds' has run or not.
        size_thresh: int
                The upper size limit for the objects that will be recognised as noise and removed from the 
                segmented image by the method 'clean_local_thresholded'.
    Methods:
    ----------------
        _calculate_local_thresholds: 
                Main method to calculate the thresholds and the thresholded, binary image.
        get_local_thresholds: 
                A method that returns local thresholds mapped to an array with same shape as 'image'.
        get_local_thresholded: 
                A method that returns the binary image thresholded with both local and global thresholds. 
        update_preliminary_mask: 
                A method to update the preliminary mask based on new inputs of 'threshold_mask' and/or 'padder'.
                After the new parameter setting, the local thresholds and thresholded volume can be automatically
                updated according to the new settings.
        update_radii:
                A method to set a new 'threshold_radii'. After the new parameter setting, the local thresholds 
                and thresholded volume are automatically updated with the new settings.
        clean_local_thresholded: 
                A method that cleans the thresholded image from small objects and border artifacts.
        update_size_thresh: 
                A method that sets a new 'size_thresh'. After the new parameter setting, the cleaning of the
                locally segmented image will be repeated with the new settings."""
    def __init__(self,image,threshold_factor=0.66,padder=None,radii=(3,9,9),size_thresh=0):
        self.img=image.copy()        
        self.threshold_factor=threshold_factor
        self.padder=padder if padder is not None else np.max(radii)
        self.radii=radii
        self.size_thresh=size_thresh
        self.padded=np.pad(self.img,(self.padder,self.padder))
        self.selected_coords,self.global_thresholded,self.global_threshold=get_preliminary_mask(self.padded,self.threshold_factor,self.padder)
        self.thresholding_status=0    
        self.local_thresholds=np.zeros_like(self.padded,dtype=float)
        self.local_thresholded=np.zeros_like(self.padded)>0
        self.local_thresholded_cleaned=self.local_thresholded.copy()
    def _calculate_local_thresholds(self):
        """This is the core method of the class. It calculates the global and local thresholds with the given
        set of the parameters and apply these thresholds to the 'img'. Each time a threshold-related parameter is
        changed, this method should be re-run to update the results with the modified parameter."""
        thresh_buffer=self.local_thresholds.copy()
        threshes=_local_otsu(self.padded,self.selected_coords,self.radii)
        thresh_buffer[tuple(self.selected_coords.T)]=threshes
        # local_thresholds=thresh_buffer[self.padder:-self.padder,self.padder:-self.padder,self.padder:-self.padder]
        self.local_thresholds=thresh_buffer.copy()
        # global_thresholded=np.pad(self.global_thresholded,(self.padder,self.padder))
        final_threshed=(self.padded>self.local_thresholds)*self.global_thresholded
        self.local_thresholded=final_threshed[self.padder:-self.padder,self.padder:-self.padder,self.padder:-self.padder]     
        self.local_thresholds=self.local_thresholds[self.padder:-self.padder,self.padder:-self.padder,self.padder:-self.padder]     
        self.thresholding_status=1
    def get_local_thresholds(self):
        """A convenience method that returns the local thresholds mapped to an array with the same shape as 'img'. 
        if the thresholds were already calculated they are directly returned. If they were not, the method calculates 
        them via '_calculate_local_thresholds' method and then returns them."""
        if self.thresholding_status:
            return(self.local_thresholds)
        else:
            self._calculate_local_thresholds()
            return(self.local_thresholds)
    def get_local_thresholded(self):
        """A convenience method that returns the final segmented image. If the segmentation was already done, the segmented 
        image is directly returned. If not, the method performs it via '_calculate_local_thresholds' method and then returns 
        the segmented image."""
        if self.thresholding_status:
            return(self.local_thresholded)
        else:
            self._calculate_local_thresholds()
            return(self.local_thresholded)
    def update_preliminary_mask(self,new_threshold_factor,new_padder=None, update_segmentation=True):
        """This is a convenience method that sets new values to the parameters 'threshold_factor' and 'padder', and then 
        updates the preliminary mask, accordingly. If any of these parameters has 'None' as input, then they will be assumed 
        to carry their current values. If the parameter update_segmentation is true, then the subsequent steps in the
        segmentation are also updated. 
        
        Parameters: 
        ----------------
            new_threshold_factor: float
                    The new value to be assigned to the parameter 'threshold_factor'. If None is assigned, the current 
                    'threshold_factor' will be used.
            new_padder: int
                    The new value to be assigned to the parameter 'padder'. Defaults to None, which means the current 'padder'
                    should be used.
            update_segmentation: bool
                    If the segmentation step should already be updated. Defaults to True."""
        self.thresholding_status=0
        if new_threshold_factor is None:
            new_threshold_factor=self.threshold_factor
        if new_padder is None:
            new_padder=self.padder
        self.threshold_factor=new_threshold_factor
        self.padder=new_padder
        self.selected_coords,self.global_thresholded,self.global_threshold=get_preliminary_mask(self.img,self.threshold_factor,self.padder)  
        if update_segmentation:
            self._calculate_local_thresholds()
    def update_radii(self,radii,update_segmentation=True):
        """This is a convenience method that sets new values to the parameter 'radii' and then updates the 
        segmentation accordingly.
        
        Parameters: 
        ----------------
            radii: (int,int,int)
                    The new values to be assigned to the parameter 'radii'. """
        self.thresholding_status=0   
        self.radii=radii
        self._calculate_local_thresholds()
    def clean_local_thresholded(self,size_thresh=100):
        """A method that cleans the thresholded image from small objects and border artifacts.
        
        Parameters: 
        ----------------
            size_thresh: int
                    The upper size limit for the objects to be removed from the segmented image."""
        radii=self.radii
        max_rad=np.max(radii)
        unpadded=self.local_thresholded[max_rad:-max_rad,max_rad:-max_rad,max_rad:-max_rad]
        repadded=np.pad(unpadded,pad_width=(max_rad,max_rad))
        cleaned=morphology.remove_small_objects(repadded,size_thresh)   
        self.local_thresholded_cleaned=cleaned.copy()
        return(self.local_thresholded_cleaned)
    def update_size_thresh(self,new_size_thresh):
        """A method to update the size threshold for the cleaning of the segmented image. 
        
        Parameters: 
        ----------------
            new_size_thresh: int
                    The new value to be assigned to the parameter 'size_thresh'. """
        self.size_thresh=new_size_thresh    
        self.local_thresholded_cleaned=self.clean_local_thresholded(self.size_thresh)
######################################################################################################################
######################################################################################################################




