# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 08:51:42 2020

@author: ozdemir
"""


""" Segmentation and quantitative comparison of the mean intensities between two sets of confocal z-stacks. 

This script can be run from a python IDE or from the command line. In order to make sure that the script can be

run from the command line, all modules and this script should be saved in a directory that is located within the PYTONPATH. 

Parameter configuration should be done directly by modifying the 'local_otsu_parameters' file. 'local_otsu_parameters_default'

holds the default parameters that were used in our analysis.


Functions:
    
    dir_path: A function that checks if any directory passed as an argument to a function is really a valid directory. 
    If not, an error is raised. 
    
    main: The main pipeline that receives images from a given directory, processes them and saves the processed images and
    the quantification results in another given directory. """
    


import numpy as np
import imageio
import os
import argparse

from local_otsu_segmentation import high_throughput_quantifier
from local_otsu_parameters import params as new_parameters
from local_otsu_parameters_default import default_params as default_parameters

def dir_path(string):
    """A function that checks if any directory passed as an argument to a function is really a valid directory. 
    If not, an error is raised. 
    
    Parameters:
    ----------------
        string: str
                The input string. If not a valid directory, an error is raised.
    Returns: 
    ----------------
        string: str
                The parameter string is simply returned if it is a valid directory."""
    if os.path.isdir(string):
        print('path_recognised')
        return string
    else:
        raise NotADirectoryError(string)

parser=argparse.ArgumentParser(description='Segmentation and Quantification of Multiple Images.')
parser.add_argument('-id','--input_directory',type=dir_path,default= None)
parser.add_argument('-od','--output_directory',type=dir_path,default=None)
### It is highly advised to specify the name list here in the parser by setting a list of names to the 'save_names' parameter. 
### Specifying the name list in the command line is not as convenient.
parser.add_argument('-sn','--save_names',type=list,default=["non_opti_0","non_opti_1","non_opti_2","opti_0","opti_1","opti_2"])
parser.add_argument('-p','--params',type=str,default='default')

args=parser.parse_args()

def main(input_directory,output_directory,save_names,parameters):
    """Parameters:
    ----------------
        input_directory: str
                The directory, where the images to be processed are located. 
        input_directory: str
                The directory, where the processed images are saved.
        save_names: list
                The main body of the file names that are given to the processed images before saving.
                This parameter is better specified in this file, by setting a default list of strings to
                the 'save_names' parameter in the argument parser. 
        parameters: str
                A string that is used to decide whether default or new parameter set will be used.
                if 'default' is passed, the default parameters will be used for the processing. If 
                'parameters' is passed, user-configured parameters will be used."""
 
    if parameters=='parameters':
        params=new_parameters
    elif parameters=='default':
        params=default_parameters
    else:
        print('only "parameters" and "default" are valid statements.')
    input_list=os.listdir(input_directory)    
    imgs=[]
    for item_in,item_out in zip(input_list,save_names):
        input_file=os.path.join(input_directory,item_in)
        npstack=imageio.volread(input_file)
        imgs.append(npstack)
    results,filtered,binaries=high_throughput_quantifier(imgs,names=save_names,params=params) 
    filter_types=['median_new.tiff', 'unsharp_mask_new.tiff','deconvolution_new.tiff','correction_new.tiff']
    core_path=output_directory
    excel_path=os.path.join(core_path,r'quantifications.xlsx')
    results.to_excel(excel_writer = excel_path)        
    for filt,binn,name in zip(filtered,binaries,save_names):
        for i,item in enumerate(filter_types):
            name_filt=os.path.join(core_path,name+'_'+item)
            filtered_=filt[i].astype(np.float32)
            imageio.volwrite(name_filt,filtered_)
        name_bin=os.path.join(core_path,name+'_'+'binary_new.tiff')
        imageio.volwrite(name_bin,binn.astype(np.uint8))


if __name__ == '__main__':
    main(args.input_directory,args.output_directory,args.save_names,args.params)
else:
    print('The import complete.')







    