# -*- coding: utf-8 -*-
"""
Created on Sun Jul  5 15:04:05 2020

@author: bugra
"""

"""Local Adaptive Segmentation.

This module performs segmentation on 3D volumetric images using various filters and local Otsu thresholding.

Functions:
    
    fit_to_window: a convenience function that pads the image borders such that its dimensions are multiples of given integers.
    
    create_shifted_images: a convenience function that shifts image frame in each dimension by given steps.
    
    correct_shifting: returns shifted images to their original frames.
    
    unsharp_mask: a function that implements the unsharp mask edge enhancement operation in an iterative manner.
    
    rl_deconvolve: Richardson-Lucy deconvolution method, as implemented by the scikit-image package.
    
    simple_deconvolver: a convenience function that implements the 'rl_deconvolve' function with averaging filter as the psf parameter.
    
    local_corrector: a function that adjusts the voxel intensities in an image to reduce sharp intensity fluctuations in the image.
    
    local_adaptive_segmentation: a function that first processes an image with certain restoration functions 
    and then segments the processed image via local_otsu thresholding.
    
    high_throughput_quantifier: a function that receives a list of 3D images as input, finds the foreground volume and calculates
    the mean intensity of the voxels within this volume for each input image."""


import numpy as np
import pandas as pd
from skimage.filters import gaussian
from scipy import ndimage as ndi
from scipy.signal import convolve
import scipy
from local_otsu import otsu

######################################################################################################################
############################################ Some auxillary functions ################################################
######################################################################################################################
def fit_to_window(img,box=(5,5,5),return_inds=False):
    """This function extends dimensions of an image such that the image can be divided into equal-sized boxes whose 
    shape is specified with the parameter 'box'. This is achieved by padding the image dimensions to sizes that are 
    multiples of the corresponding dimensions of the given box.
    
    Parameters:
    ----------------
        image: array 
                The input data, which is padded with zero-valued voxels.
        box: (int,int,int)
                Shape of the box, which is used to calculate the pad width.
        return_inds: 
                Whether the pad widths to extend image dimensions are returned. Defaults to False.
    
    Returns: 
    ----------------
        im: array 
                Padded image that can be divided into an integer number of boxes specified with the parameter 'box'.
        remns: list
                Pad widths to extend each direction. Returned only when 'return_inds' parameter is True."""
    box=np.array(box)
    im=img.copy()
    imshape=np.array(im.shape)
    imshapeplus=imshape+box
    first_remns=imshapeplus%box
    imshapegood=imshapeplus-first_remns
    remnants=imshapegood-imshape
    remnants=remnants%box
    remnants=remnants.astype(np.int64)
    remn0=remnants//2
    remn1=remnants-remn0
    remns=[tuple(item) for item in np.array((remn0,remn1)).T]
    im=np.pad(im,pad_width=remns,mode='edge')
    if return_inds:
        return(im,remns)
    else:
        return(im)
    
def create_shifted_images(img, padder=2):
    """This function creates, from one image, 3 images that are frame-shifted from the original image by
    varying extents.
    
    Parameters:
    ----------------
        img: array 
                The input data, from which frame-shifted images are created.
        padder: int 
                Range of the frame-shift. The largest number of voxels shifted. 

    Returns: 
    ----------------
        im0: array 
                frame-shifted image no 1.
        im1: array
                frame-shifted image no 2.
        im2: array
                frame-shifted image no 3."""
    if padder<2:
        print('Padder must be at least 2.')
        return()
    padded=np.pad(img,padder,mode='edge')
    im0=padded[0:-2*padder,0:-2*padder,0:-2*padder]
    im1=padded[padder//2:-(2*padder-(padder//2)),
               padder//2:-(2*padder-(padder//2)),
               padder//2:-(2*padder-(padder//2))]
    im2=padded[(padder+padder//2):-(2*padder-(padder+padder//2)),
               (padder+padder//2):-(2*padder-(padder+padder//2)),
               (padder+padder//2):-(2*padder-(padder+padder//2))]
    return(im0,im1,im2)


def correct_shifting(shifted,padder=2):
    """This function reverses the frame-shifting induced by the 'create_shifted_images' function.
    
    Parameters:
    ----------------
        shifted: (array, array, array) 
                A list of 3 volumes, each representing a different frame-shift produced by the 
                function 'create_shifted_images' in this file.
        padder: int 
                Range of the frame-shift. This parameter must be equal to the 'padder' parameter
                that was used to create the 3 input volumes via 'create_shifted_images' function. 

    Returns: 
    ----------------
        centralised0: array 
                frame-corrected image no 1.
        centralised1: array
                frame-corrected image no 2.
        centralised2: array
                frame-corrected image no 3."""
    im0,im1,im2=shifted
    padded0=np.pad(im0,padder,mode='edge')
    padded1=np.pad(im1,padder,mode='edge')
    padded2=np.pad(im2,padder,mode='edge')
    centralised0=padded0[2*padder:,2*padder:,2*padder:]
    centralised1=padded1[(2*padder-(padder//2)):-(padder//2),
                         (2*padder-(padder//2)):-(padder//2),
                         (2*padder-(padder//2)):-(padder//2)]
    centralised2=padded2[(2*padder-(padder+padder//2)):-(padder+padder//2),
                         (2*padder-(padder+padder//2)):-(padder+padder//2),
                         (2*padder-(padder+padder//2)):-(padder+padder//2)]
    return(centralised0,centralised1,centralised2)
######################################################################################################################
######################################################################################################################



######################################################################################################################
############################################ Image restoration functions #############################################
######################################################################################################################
def unsharp_mask(img,n=1,iterations=3,sig=(1,1,1)): 
    """This function enhances the edge-contrast in the images by using unsharp mask filter. 
    This filter subtracts a Gaussian-blurred form of an image from the image itself and
    then adds this difference back to the original image in an iterative manner.
    
    Parameters:
    ----------------
        img: array
                The input data, which is subjected to unsharp-mask filter.
        n:  float
                A factor that controls the degree of the sharpening. The greater the n, the greater the
                sharpening effect at each iteration. Too high n can lead to the loss of signal at non-edge 
                regions.
        iterations: int
                The number of times the unsharp mask filter is iterated
        sig: tuple of floats
                The sigma parameter of the Gaussian blur in each dimension. Must be of the length that matches
                the number dimensions of 'img'.

    Returns: 
    ----------------
        centralised0: array 
                frame-corrected image no 1.
        centralised1: array
                frame-corrected image no 2.
        centralised2: array
                frame-corrected image no 3."""
    im=img/img.max()
    eps=np.finfo(im.dtype).eps
    for i in range(iterations):
        gim=gaussian(im,sig)
        im=im+(im-gim)*n
        im[im<=0]=eps
        im=im/im.max()
    return(im)

def rl_deconvolve(image, psf, iterations=5):
    """Richardson-Lucy deconvolution as implemented in the scikit-image package.
    This code was modified from the Richardson Lucy deconvolution function in the scikit-image
    package.
    
    Parameters:
    ----------------
        image: array
                The input data to be deconvolved.
        psf: array
                point spread function for the deconvolution process. Must have the same 
                number of dimensions as the 'image'.
        iterations: int
                The number of times the deconvolution is iterated

    Returns: 
    ----------------
        im_deconv: array of dtype float
                The deconvolved image."""
    image = image.astype(np.float)
    eps=np.finfo(image.dtype).eps
    psf = psf.astype(np.float)
    im_deconv = np.full(image.shape, 0.5)
    psf_mirror = psf[::-1, ::-1, ::-1]
    conv_im=convolve(im_deconv, psf_mirror, mode='same') 
    for _ in range(iterations):
        np.place(conv_im,conv_im==0,eps)
        relative_blur = image / conv_im + eps
        c=convolve(relative_blur, psf, mode='same') 
        im_deconv *= c
        conv_im=convolve(im_deconv, psf_mirror, mode='same') 
    return(im_deconv) 

def simple_deconvolver(img,psf_shape=(5,5,5),iterations=5):
    """A convenience function that applies the 'rl_deconvolve' function using an averaging filter
    of given shape as the point spread function. The aim of this function is not a true deblurring, 
    but rather a smoothing effect without losing too much detail in structure. 
    
    Parameters:
    ----------------
        img: array
                The input data, which is subjected to unsharp-mask operation.
        psf_shape: tuple of integers
                the shape, with which an averaging filter is created as the point 
                spread function for the deconvolution process.
        iterations: int
                The number of times the deconvolution is iterated

    Returns: 
    ----------------
        im_deconv: array of dtype float
                The deconvolved image."""
    im=img/img.max()
    psf=np.ones(psf_shape)/np.prod(psf_shape)
    eps=np.finfo(im.dtype).eps
    if iterations>0:
        deconvolved=rl_deconvolve(im,psf,iterations)
    else:
        deconvolved=im.copy()
    deconvolved[deconvolved<=0]=eps
    return(deconvolved)

def local_corrector(img, window=(10,10,10), blur_fact=1, power=1, kfact=1, init_sig=0, grad_sig=0):
    """A function that reduces sharp intensity fluctuations in the image in order to facilitate
    the segmentation process. This is done by first dividing the image into equal-shaped boxes. 
    Each box is then separately processed by a function, which uses the box histogram to readjust 
    the voxel intensities. 
    
    Parameters:
    ----------------
        img: array
                The input data, to which the function is applied. Must be 3-dimensional.
        window: tuple of integers
                The shape of the box, which is one of the equal-sized pieces the 'img' is 
                divided into. Histograms of the boxes are used to develop local multipliers 
                that readjust the voxel intensities of the original image.
        blur_fact: float
                sigma factor for the Gaussian filter that is used to make the box transitions 
                more continuous and reduce possible artefacts at the box boundaries.
        power: float
                Power that is applied to the histogram modification factor before multiplying 
                the image with it. 
        kfact: float
                A parameter that controls the extent of the histogram modification. Higher
                values lead to milder changes in the image. 
        init_sig: float
                The sigma parameter for an initial gaussian blurring of the image before the actual
                processing starts. 
        grad_sig: float
                The sigma parameter for the calculation of the image gradient magnitude. If this parameter
                is non-zero, then the gradient magnitude image will be used for determining the histogram 
                modification factor. 
    Returns: 
    ----------------
        finalised: array of dtype float
                The processed image."""
    im=img/img.max()
    eps=np.finfo(im.dtype).eps
    im[im==0]=eps
    padded,remns=fit_to_window(im,window,return_inds=True)
    remns=np.array(remns)
    lowbounds=remns[:,0]
    highbounds=np.array(padded.shape)-remns[:,1]
    padded=gaussian(padded,init_sig)
    if grad_sig==0:
        hist_template=padded.copy()
    else:
        hist_template=ndi.filters.gaussian_gradient_magnitude(padded,grad_sig)
    hist_template[hist_template<0]=eps
    dlim,rlim,clim=padded.shape
    drange=np.arange(0,dlim+1,window[0])    
    rrange=np.arange(0,rlim+1,window[1])
    crange=np.arange(0,clim+1,window[2])
    skewlist=[]
    boxlist=[]
    for i,d in enumerate(drange[1:]):
        dmin=drange[i]
        for j,r in enumerate(rrange[1:]):
            rmin=rrange[j]
            for k,c in enumerate(crange[1:]):
                cmin=crange[k]
                box=hist_template[dmin:d+1,rmin:r+1,cmin:c+1]
                skw=scipy.stats.skew(box.flatten())
                boxlist.append(box)
                skewlist.append(skw)
    skewlist=np.array(skewlist)
    skewlist[np.isnan(skewlist)]=eps
    skewlist[skewlist==0]=eps
    skewlist[np.isinf(skewlist)]=0
    skewlist[skewlist==0]=np.max(skewlist)
    buffer=np.zeros_like(padded)
    l=0
    for i,d in enumerate(drange[1:]):
        dmin=drange[i]
        for j,r in enumerate(rrange[1:]):
            rmin=rrange[j]
            for k,c in enumerate(crange[1:]):
                cmin=crange[k]
                box=buffer[dmin:d+1,rmin:r+1,cmin:c+1]    
                box[box>-5]=skewlist[l]
                buffer[dmin:d+1,rmin:r+1,cmin:c+1]=box
                l=l+1
    factor=buffer/np.max(np.abs(buffer))
    factor=factor-factor.min()
    factor[factor==0]=eps
    factor=(gaussian(factor,blur_fact)+kfact)**power
    factor=factor/np.max(np.abs(factor))
    modified=padded*factor
    modified=modified/np.max(np.abs(modified))
    finalised=modified[lowbounds[0]:highbounds[0],lowbounds[1]:highbounds[1],lowbounds[2]:highbounds[2]]
    return(finalised)
######################################################################################################################
######################################################################################################################


######################################################################################################################
################################ Main filtering and segmentation pipeline ############################################
######################################################################################################################


def local_adaptive_segmentation(img,params,shift_scale=2,return_filtered=False):
    """The main binarisation pipeline that receives an input image, pre-processes it through a series
    of functions in preparation for the thresholding and finally applies an adaptive thresholding 
    to the processed image. 
    
    Parameters:
    ----------------
        img: array
                The input data, which is processed through the pipeline. Must be 3-dimensional.
        params: dict
                The collection of parameters that can be used to tune multiple steps of the pipeline.
                These are the reachable and modifiable parameters from the 'main.py' file. 
                Accessed with the following keys: 
                    'median_filter': [(int,int,int)]
                        Shape of the local window for a median filter.
                    'unsharp_mask': [float,int,(float,float,float)],
                        Parameters of the unsharp_mask function: 'n', 'iterations', 'sig'.
                    'Richardson_Lucy': [(int,int,int), int]
                        Parameters of the 'simple_deconvolver' function: 'psf_shape', 'iterations'.
                    'spatial_correction': [(int,int,int), float, float, float, float, float].
                        Parameters of the 'local_corrector' function: 'window','blur_fact','power','kfact','init_sig','grad_sig'
                    'segmentation': [float,int,(int,int,int)]
                        Parameters for the __init__ method of the 'otsu' class: 'threshold_factor', 'padder', 'radii'        
        shift_scale: int
                Range of the frame-shift that is induced on the input image before applying the 'local_corrector'
                function. The idea behind the frame-shifting is that the 'local_corrector' function samples the image
                in three different frames. Taking the mean of the three outputs reduces possible artefacts that may 
                occur at the borders of the boxes. 
        return_filtered:
                If True, returns the outputs of the intermediary functions in addition to the segmented image. 
                If False, only the segmented image is returned.  
    Returns: 
    ----------------
        *denoised: Output of the median filter.
        *edge_enhanced: Output of the 'unsharp_mask function'.
        *decon: Output of the 'simple_deconvolver' function.
        *mean_img: The mean of the three outputs of the 'local_corrector' function with alternative box frames.
        seg: The output of the segmentation process via the 'otsu' class.
        
        Those marked with * are returned only if the parameter 'return_filtered' is True."""
        
    im=img/img.max()
    denoised=ndi.median_filter(im,footprint=np.ones(params['median_filter'][0]))
    edge_enhanced=unsharp_mask(denoised,*params['unsharp_mask'])
    print('denoising_and_edge_enhancement_finished')
    decon=simple_deconvolver(edge_enhanced, *params['Richardson_Lucy'])
    print('Richardson_Lucy_restoration_finished')
    decon_shifted=create_shifted_images(decon,shift_scale)
    corrected=[]
    for item in decon_shifted:
        decon_cor=local_corrector(decon,*params['spatial_correction'])    
        decon_cor_blur=decon_cor.copy()
        corrected.append(decon_cor_blur)
        print('spatial_correction_finished')
    recentered=correct_shifting(corrected,shift_scale)
    mean_img=np.array(recentered).mean(axis=0)
    segger=otsu(mean_img,*params['segmentation'])
    seg=segger.get_local_thresholded()
    print('local_adaptive_segmentation_finished')
    if return_filtered:
        return(denoised,edge_enhanced,decon,mean_img,seg)
    else:
        return(seg)
    

def high_throughput_quantifier(imset,names,params):
    """This is a function that performs image processing, segmentation and quantification of multiple images with a
    given set of parameters. The quantification involves the total voxel numbers in the binary images, total voxel intensity
    within the volume limited by the binary mask, mean voxel intensity within the volume limited by the binary mask.
    
    Parameters:
    ----------------
        imset: list
                The input data, which consists of multiple 3D arrays.
        names: list
                A list of strings that are used as names of the input volumes. These names appear in the returned DataFrame,
                which contains the quantification results. 
        params: dict
                The collection of parameters that can be used to tune multiple steps of the pipeline.
                These are the reachable and modifiable parameters from the 'main.py' file. 
                Accessed with the following keys: 
                    'median_filter': [(int,int,int)]
                        Shape of the local window for a median filter.
                    'unsharp_mask': [float,int,(float,float,float)],
                        Parameters of the unsharp_mask function: 'n', 'iterations', 'sig'.
                    'Richardson_Lucy': [(int,int,int), int]
                        Parameters of the 'simple_deconvolver' function: 'psf_shape', 'iterations'.
                    'spatial_correction': [(int,int,int), float, float, float, float, float].
                        Parameters of the 'local_corrector' function: 'window','blur_fact','power','kfact','init_sig','grad_sig'
                    'segmentation': [float,int,(int,int,int)]
                        Parameters for the __init__ method of the 'otsu' class: 'threshold_factor', 'padder', 'radii'        
    Returns: 
    ----------------
        df: DataFrame
                A DataFrame that contains the following quantitative information concerning the input images: 'total_volume',
                'total_intensity', 'mean_intensity'. There is also another column with the title 'names', which stores the labels
                given to the respective images.
        filtered_series: list
                A list of tuples, each tuple containing 4 outputs from the intermediary functions: 'median_filter',
                'unsharp_mask', 'simple_deconvolver' and 'local_corrector'.
        segmented_series: list
                A list of arrays, each array being the binary mask yielded by the segmentation process. This binary mask is
                used to select the voxels in the original image to perform the quantitative analysis."""
    database=[]
    segmented_series,filtered_series=[],[]
    for i, item in enumerate(imset):
        all_filtered=local_adaptive_segmentation(item,params,shift_scale=2,return_filtered=True)
        segmented=all_filtered[-1].copy()
        filtered=all_filtered[:-1]
        segmented_series.append(segmented)
        filtered_series.append(filtered)
        masked=segmented*imset[i]
        total_volume=segmented.sum()
        total_intensity=masked.sum()
        mean_intensity=total_intensity/total_volume
        database.append([total_volume,total_intensity,mean_intensity])
        print('image_name: '+names[i])
        print('image_no:{0},image_volume:{1},image_total_intensity:{2},image_mean_intensity:{3}'.format(i,total_volume,total_intensity,mean_intensity))
    df=pd.DataFrame(database,columns=['total_volume','total_intensity','mean_intensity'])
    df=pd.concat((df,pd.Series(names,name='names')),axis=1)
    return(df,filtered_series,segmented_series)

######################################################################################################################
######################################################################################################################
